from waveforms.qlisp.macro import (_VZ_P, _VZ_clear, _VZ_exchangable,
                                   _VZ_rfUnitary, _VZ_swap, add_VZ_rule,
                                   exchangeRzWithGate, extend_macro,
                                   reduceVirtualZ, remove_VZ_rule)
